"""Generates the supported SOP Classes."""

import inspect
import logging
import sys

from pydicom.uid import UID

from pynetdicom.service_class import (
    ServiceClass,
    VerificationServiceClass,
    StorageServiceClass,
    QueryRetrieveServiceClass,
    BasicWorklistManagementServiceClass,
    RelevantPatientInformationQueryServiceClass,
    SubstanceAdministrationQueryServiceClass,
    NonPatientObjectStorageServiceClass,
    HangingProtocolQueryRetrieveServiceClass,
    DefinedProcedureProtocolQueryRetrieveServiceClass,
    ColorPaletteQueryRetrieveServiceClass,
    ImplantTemplateQueryRetrieveServiceClass,
)
from pynetdicom.service_class_n import (
    DisplaySystemManagementServiceClass
)


LOGGER = logging.getLogger('pynetdicom.sop')


def uid_to_service_class(uid):
    """Return the ServiceClass object corresponding to `uid`.

    Parameters
    ----------
    uid : pydicom.uid.UID
        The SOP or Service Class UID to use to find the corresponding Service
        Class.

    Returns
    -------
    service_class.ServiceClass
        The Service Class corresponding to the SOP Class UID.

    Raises
    ------
    NotImplementedError
        If the Service Class corresponding to the SOP Class `uid` hasn't been
        implemented.
    """
    if uid in _SERVICE_CLASSES:
        return _SERVICE_CLASSES[uid]
    elif uid in _VERIFICATION_CLASSES.values():
        return VerificationServiceClass
    elif uid in _STORAGE_CLASSES.values():
        return StorageServiceClass
    elif uid in _QR_CLASSES.values():
        return QueryRetrieveServiceClass
    elif uid in _BASIC_WORKLIST_CLASSES.values():
        return BasicWorklistManagementServiceClass
    elif uid in _RELEVANT_PATIENT_QUERY_CLASSES.values():
        return RelevantPatientInformationQueryServiceClass
    elif uid in _SUBSTANCE_ADMINISTRATION_CLASSES.values():
        return SubstanceAdministrationQueryServiceClass
    elif uid in _NON_PATIENT_OBJECT_CLASSES.values():
        return NonPatientObjectStorageServiceClass
    elif uid in _HANGING_PROTOCOL_CLASSES.values():
        return HangingProtocolQueryRetrieveServiceClass
    elif uid in _DEFINED_PROCEDURE_CLASSES.values():
        return DefinedProcedureProtocolQueryRetrieveServiceClass
    elif uid in _COLOR_PALETTE_CLASSES.values():
        return ColorPaletteQueryRetrieveServiceClass
    elif uid in _IMPLANT_TEMPLATE_CLASSES.values():
        return ImplantTemplateQueryRetrieveServiceClass
    elif uid in _DISPLAY_SYSTEM_CLASSES.values():
        return DisplaySystemManagementServiceClass
    elif uid in _PRINT_MANAGEMENT_CLASSES.values():
        return ServiceClass  # Not yet implemented
    elif uid in _PROCEDURE_STEP_CLASSES.values():
        return ServiceClass  # Not yet implemented
    elif uid in _MEDIA_STORAGE_CLASSES.values():
        return ServiceClass  # Not yet implemented
    elif uid in _UNIFIED_PROCEDURE_STEP_CLASSES.values():
        return ServiceClass  # Not yet implemented
    elif uid in _RT_MACHINE_VERIFICATION_CLASSES.values():
        return ServiceClass  # Not yet implemented

    # No SCP implemented
    return ServiceClass


class SOPClass(UID):
    """Extend pydicom's UID to include the corresponding Service Class."""
    _service_class = None

    def __new__(cls, val):
        if isinstance(val, SOPClass):
            return val

        return super(SOPClass, cls).__new__(cls, val)

    def __getattribute__(self, name):
        return super(SOPClass, self).__getattribute__(name)

    @property
    def service_class(self):
        """Return the corresponding Service Class implementation."""
        return self._service_class


def _generate_sop_classes(sop_class_dict):
    """Generate the SOP Classes."""
    for name in sop_class_dict:
        sop_class = SOPClass(sop_class_dict[name])
        sop_class._service_class = uid_to_service_class(sop_class_dict[name])
        globals()[name] = sop_class


# Table of service classes with assigned UIDs
_SERVICE_CLASSES = {
    '1.2.840.10008.4.2' : StorageServiceClass,
    #'1.2.840.10008.5.1.4.34.6', UnifiedProcedureStepServiceClass,
}

# Generate the various SOP classes
_VERIFICATION_CLASSES = {
    'VerificationSOPClass' : '1.2.840.10008.1.1',
}

# pylint: disable=line-too-long
_STORAGE_CLASSES = {
    'ComputedRadiographyImageStorage' : '1.2.840.10008.5.1.4.1.1.1',  # A.2
    'DigitalXRayImagePresentationStorage' : '1.2.840.10008.5.1.4.1.1.1.1',  # A.26
    'DigitalXRayImageProcessingStorage' : '1.2.840.10008.5.1.4.1.1.1.1.1',  # A.26
    'DigitalMammographyXRayImagePresentationStorage' : '1.2.840.10008.5.1.4.1.1.1.2',  # A.27
    'DigitalMammographyXRayImageProcessingStorage' : '1.2.840.10008.5.1.4.1.1.1.2.1',  # A.27
    'DigitalIntraOralXRayImagePresentationStorage' : '1.2.840.10008.5.1.4.1.1.1.3',  # A.28
    'DigitalIntraOralXRayImageProcessingStorage' : '1.2.840.10008.5.1.4.1.1.1.3.1',  # A.28
    'CTImageStorage' : '1.2.840.10008.5.1.4.1.1.2',  # A.3
    'EnhancedCTImageStorage' : '1.2.840.10008.5.1.4.1.1.2.1',  # A.38
    'LegacyConvertedEnhancedCTImageStorage' : '1.2.840.10008.5.1.4.1.1.2.2',  # A.70
    'UltrasoundMultiframeImageStorage' : '1.2.840.10008.5.1.4.1.1.3.1',  # A.7
    'MRImageStorage' : '1.2.840.10008.5.1.4.1.1.4',  # A.4
    'EnhancedMRImageStorage' : '1.2.840.10008.5.1.4.1.1.4.1',  # A.36.2
    'MRSpectroscopyStorage' : '1.2.840.10008.5.1.4.1.1.4.2',  # A.36.3
    'EnhancedMRColorImageStorage' : '1.2.840.10008.5.1.4.1.1.4.3',  # A.36.4
    'LegacyConvertedEnhancedMRImageStorage' : '1.2.840.10008.5.1.4.1.1.4.4',  # A.71
    'UltrasoundImageStorage' : '1.2.840.10008.5.1.4.1.1.6.1',  # A.6
    'EnhancedUSVolumeStorage' : '1.2.840.10008.5.1.4.1.1.6.2',  # A.59
    'SecondaryCaptureImageStorage' : '1.2.840.10008.5.1.4.1.1.7',  # A.8.1
    'MultiframeSingleBitSecondaryCaptureImageStorage' : '1.2.840.10008.5.1.4.1.1.7.1',  # A.8.2
    'MultiframeGrayscaleByteSecondaryCaptureImageStorage' : '1.2.840.10008.5.1.4.1.1.7.2',  # A.8.3
    'MultiframeGrayscaleWordSecondaryCaptureImageStorage' : '1.2.840.10008.5.1.4.1.1.7.3',  # A.8.4
    'MultiframeTrueColorSecondaryCaptureImageStorage' : '1.2.840.10008.5.1.4.1.1.7.4',  # A.8.5
    'TwelveLeadECGWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.1.1',  # A.34.3
    'GeneralECGWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.1.2',  # A.34.4
    'AmbulatoryECGWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.1.3',  # A.34.5
    'HemodynamicWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.2.1',  # A.34.6
    'CardiacElectrophysiologyWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.3.1',  # A.34.7
    'BasicVoiceAudioWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.4.1',  # A.34.2
    'GeneralAudioWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.4.2',  # A.34.10
    'ArterialPulseWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.5.1',  # A.34.8
    'RespiratoryWaveformStorage' : '1.2.840.10008.5.1.4.1.1.9.6.1',  # A.34.9
    'GrayscaleSoftcopyPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.1',  # A.33.1
    'ColorSoftcopyPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.2',  # A.33.2
    'PseudocolorSoftcopyPresentationStageStorage' : '1.2.840.10008.5.1.4.1.1.11.3',  # A.33.3
    'BlendingSoftcopyPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.4',  # A.33.4
    'XAXRFGrayscaleSoftcopyPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.5',  # A.33.6
    'GrayscalePlanarMPRVolumetricPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.6',  # A.80.1
    'CompositingPlanarMPRVolumetricPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.7',  # A.80.1
    'AdvancedBlendingPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.8',  # A.33.7
    'VolumeRenderingVolumetricPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.9',  # A.80.2
    'SegmentedVolumeRenderingVolumetricPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.10',  # A.80.2
    'MultipleVolumeRenderingVolumetricPresentationStateStorage' : '1.2.840.10008.5.1.4.1.1.11.11',  # A.80.2
    'XRayAngiographicImageStorage' : '1.2.840.10008.5.1.4.1.1.12.1',  # A.14
    'EnhancedXAImageStorage' : '1.2.840.10008.5.1.4.1.1.12.1.1',  # A.47
    'XRayRadiofluoroscopicImageStorage' : '1.2.840.10008.5.1.4.1.1.12.2',  # A.16
    'EnhancedXRFImageStorage' : '1.2.840.10008.5.1.4.1.1.12.2.1',  # A.48
    'XRay3DAngiographicImageStorage' : '1.2.840.10008.5.1.4.1.1.13.1.1',  # A.53
    'XRay3DCraniofacialImageStorage' : '1.2.840.10008.5.1.4.1.1.13.1.2',  # A.54
    'BreastTomosynthesisImageStorage' : '1.2.840.10008.5.1.4.1.1.13.1.3',  # A.55
    'BreastProjectionXRayImagePresentationStorage' : '1.2.840.10008.5.1.4.1.1.13.1.4',  # A.74
    'BreastProjectionXRayImageProcessingStorage' : '1.2.840.10008.5.1.4.1.1.13.1.5',  # A.74
    'IntravascularOpticalCoherenceTomographyImagePresentationStorage' : '1.2.840.10008.5.1.4.1.1.14.1',  # A.66
    'IntravascularOpticalCoherenceTomographyImageProcessingStorage' : '1.2.840.10008.5.1.4.1.1.14.2',  # A.66
    'NuclearMedicineImageStorage' : '1.2.840.10008.5.1.4.1.1.20',    # A.5
    'ParametricMapStorage' : '1.2.840.10008.5.1.4.1.1.30',  # A.75
    'RawDataStorage' : '1.2.840.10008.5.1.4.1.1.66',  # A.37
    'SpatialRegistrationStorage' : '1.2.840.10008.5.1.4.1.1.66.1',  # A.39.1
    'SpatialFiducialsStorage' : '1.2.840.10008.5.1.4.1.1.66.2',  # A.40
    'DeformableSpatialRegistrationStorage' : '1.2.840.10008.5.1.4.1.1.66.3',  # A.39.2
    'SegmentationStorage' : '1.2.840.10008.5.1.4.1.1.66.4',  # A.51
    'SurfaceSegmentationStorage' : '1.2.840.10008.5.1.4.1.1.66.5',  # A.57
    'TractographyResultsStorage' : '1.2.840.10008.5.1.4.1.1.66.6',  # A.78
    'RealWorldValueMappingStorage' : '1.2.840.10008.5.1.4.1.1.67',  # A.46
    'SurfaceScanMeshStorage' : '1.2.840.10008.5.1.4.1.1.68.1',  # A.68
    'SurfaceScanPointCloudStorage' : '1.2.840.10008.5.1.4.1.1.68.2',  # A.69
    'VLEndoscopicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.1',  # A.32.1
    'VideoEndoscopicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.1.1',  # A.32.5
    'VLMicroscopicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.2',  # A.32.2
    'VideoMicroscopicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.2.1',  # A.32.6
    'VLSlideCoordinatesMicroscopicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.3',  # A.32.3
    'VLPhotographicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.4',  # A.32.4
    'VideoPhotographicImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.4.1',  # A.32.7
    'OphthalmicPhotography8BitImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.1',  # A.41
    'OphthalmicPhotography16BitImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.2',  # A.42
    'StereometricRelationshipStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.3',  # A.43
    'OphthalmicTomographyImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.4',  # A.52
    'WideFieldOphthalmicPhotographyStereographicProjectionImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.5',  # A.76
    'WideFieldOphthalmicPhotography3DCoordinatesImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.6',  # A.77
    'OphthalmicOpticalCoherenceTomographyEnFaceImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.7',  # A.83
    'OphthlamicOpticalCoherenceTomographyBScanVolumeAnalysisStorage' : '1.2.840.10008.5.1.4.1.1.77.1.5.8',  # A.84
    'VLWholeSlideMicroscopyImageStorage' : '1.2.840.10008.5.1.4.1.1.77.1.6',  # A.32.8
    'LensometryMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.78.1',  # A.60.1
    'AutorefractionMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.78.2',  # A.60.2
    'KeratometryMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.78.3',  # A.60.3
    'SubjectiveRefractionMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.78.4',  # A.60.4
    'VisualAcuityMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.78.5',  # A.60.5
    'SpectaclePrescriptionReportStorage' : '1.2.840.10008.5.1.4.1.1.78.6',  # A.35.9
    'OphthalmicAxialMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.78.7',  # A.60.6
    'IntraocularLensCalculationsStorage' : '1.2.840.10008.5.1.4.1.1.78.8',  # A.60.7
    'MacularGridThicknessAndVolumeReport' : '1.2.840.10008.5.1.4.1.1.79.1',  # A.35.11
    'OphthalmicVisualFieldStaticPerimetryMeasurementsStorage' : '1.2.840.10008.5.1.4.1.1.80.1',  # A.65
    'OphthalmicThicknessMapStorage' : '1.2.840.10008.5.1.4.1.1.81.1',  # A.67
    'CornealTopographyMapStorage' : '1.2.840.10008.5.1.4.1.1.82.1',  # A.73
    'BasicTextSRStorage' : '1.2.840.10008.5.1.4.1.1.88.11',  # A.35.1
    'EnhancedSRStorage' : '1.2.840.10008.5.1.4.1.1.88.22',  # A.35.2
    'ComprehensiveSRStorage' : '1.2.840.10008.5.1.4.1.1.88.33',  # A.35.3
    'Comprehensive3DSRStorage' : '1.2.840.10008.5.1.4.1.1.88.34',  # A.35.13
    'ExtensibleSRStorage' : '1.2.840.10008.5.1.4.1.1.88.35',  # A.35.15
    'ProcedureSRStorage' : '1.2.840.10008.5.1.4.1.1.88.40',  # A.35.7
    'MammographyCADSRStorage' : '1.2.840.10008.5.1.4.1.1.88.50',  # A.35.5
    'KeyObjectSelectionStorage' : '1.2.840.10008.5.1.4.1.1.88.59',  # A.35.4
    'ChestCADSRStorage' : '1.2.840.10008.5.1.4.1.1.88.65',  # A.35.6
    'XRayRadiationDoseSRStorage' : '1.2.840.10008.5.1.4.1.1.88.67',  # A.35.8
    'RadiopharmaceuticalRadiationDoseSRStorage' : '1.2.840.10008.5.1.4.1.1.88.68',  # A.35.14
    'ColonCADSRStorage' : '1.2.840.10008.5.1.4.1.1.88.69',  # A.35.10
    'ImplantationPlanSRDocumentStorage' : '1.2.840.10008.5.1.4.1.1.88.70',  # A.35.12
    'AcquisitionContextSRStorage' : '1.2.840.10008.5.1.4.1.1.88.71',  # A.35.16
    'SimplifiedAdultEchoSRStorage' : '1.2.840.10008.5.1.4.1.1.88.72',  # A.35.17
    'PatientRadiationDoseSRStorage' : '1.2.840.10008.5.1.4.1.1.88.73',  # A.35.18
    'ContentAssessmentResultsStorage' : '1.2.840.10008.5.1.4.1.1.90.1',  # A.81
    'EncapsulatedPDFStorage' : '1.2.840.10008.5.1.4.1.1.104.1',  # A.45.1
    'EncapsulatedCDAStorage' : '1.2.840.10008.5.1.4.1.1.104.2',  # A.45.2
    'EncapsulatedSTLStorage' : '1.2.840.10008.5.1.4.1.1.104.3',  # A.85.1
    'PositronEmissionTomographyImageStorage' : '1.2.840.10008.5.1.4.1.1.128',  # A.21
    'EnhancedPETImageStorage' : '1.2.840.10008.5.1.4.1.1.130',  # A.56
    'LegacyConvertedEnhancedPETImageStorage' : '1.2.840.10008.5.1.4.1.1.128.1',  # A.72
    'BasicStructuredDisplayStorage' : '1.2.840.10008.5.1.4.1.1.131',  # A.33.5
    'CTPerformedProcedureProtocolStorage' : '1.2.840.10008.5.1.4.1.1.200.2',  # A.82.1
    'RTImageStorage' : '1.2.840.10008.5.1.4.1.1.481.1',  # A.17
    'RTDoseStorage' : '1.2.840.10008.5.1.4.1.1.481.2',  # A.18
    'RTStructureSetStorage' : '1.2.840.10008.5.1.4.1.1.481.3',  # A.19
    'RTBeamsTreatmentRecordStorage' : '1.2.840.10008.5.1.4.1.1.481.4',  # A.29
    'RTPlanStorage' : '1.2.840.10008.5.1.4.1.1.481.5',  # A.20
    'RTBrachyTreatmentRecordStorage' : '1.2.840.10008.5.1.4.1.1.481.6',  # A.20
    'RTTreatmentSummaryRecordStorage' : '1.2.840.10008.5.1.4.1.1.481.7',  # A.31
    'RTIonPlanStorage' : '1.2.840.10008.5.1.4.1.1.481.8',  # A.49
    'RTIonBeamsTreatmentRecordStorage' : '1.2.840.10008.5.1.4.1.1.481.9',  # A.50
    'RTBeamsDeliveryInstructionStorage' : '1.2.840.10008.5.1.4.34.7',  # A.64
    'RTBrachyApplicationSetupDeliveryInstructionsStorage' : '1.2.840.10008.5.1.4.34.10',  # A.79
}

_QR_CLASSES = {
    'PatientRootQueryRetrieveInformationModelFind' : '1.2.840.10008.5.1.4.1.2.1.1',
    'PatientRootQueryRetrieveInformationModelMove' : '1.2.840.10008.5.1.4.1.2.1.2',
    'PatientRootQueryRetrieveInformationModelGet' : '1.2.840.10008.5.1.4.1.2.1.3',
    'StudyRootQueryRetrieveInformationModelFind' : '1.2.840.10008.5.1.4.1.2.2.1',
    'StudyRootQueryRetrieveInformationModelMove' : '1.2.840.10008.5.1.4.1.2.2.2',
    'StudyRootQueryRetrieveInformationModelGet' : '1.2.840.10008.5.1.4.1.2.2.3',
    'PatientStudyOnlyQueryRetrieveInformationModelFind' : '1.2.840.10008.5.1.4.1.2.3.1',
    'PatientStudyOnlyQueryRetrieveInformationModelMove' : '1.2.840.10008.5.1.4.1.2.3.2',
    'PatientStudyOnlyQueryRetrieveInformationModelGet' : '1.2.840.10008.5.1.4.1.2.3.3',
    'CompositeInstanceRootRetrieveMove' : '1.2.840.10008.5.1.4.1.2.4.2',
    'CompositeInstanceRootRetrieveGet' : '1.2.840.10008.5.1.4.1.2.4.3',
    'CompositeInstanceRetrieveWithoutBulkDataGet' : '1.2.840.10008.5.1.4.1.2.5.3',
}

_BASIC_WORKLIST_CLASSES = {
    'ModalityWorklistInformationFind' : '1.2.840.10008.5.1.4.31',
}

_RELEVANT_PATIENT_QUERY_CLASSES = {
    'GeneralRelevantPatientInformationQuery' : '1.2.840.10008.5.1.4.37.1',
    'BreastImagingRelevantPatientInformationQuery' : '1.2.840.10008.5.1.4.37.2',
    'CardiacRelevantPatientInformationQuery' : '1.2.840.10008.5.1.4.37.3',
}

_SUBSTANCE_ADMINISTRATION_CLASSES = {
    'ProductCharacteristicsQueryInformationModelFind' : '1.2.840.10008.5.1.4.41',
    'SubstanceApprovalQueryInformationModelFind' : '1.2.840.10008.5.1.4.42'
}

_NON_PATIENT_OBJECT_CLASSES = {
    'HangingProtocolStorage' : '1.2.840.10008.5.1.4.38.1',
    'ColorPaletteStorage' : '1.2.840.10008.5.1.4.39.1',
    'GenericImplantTemplateStorage' : '1.2.840.10008.5.1.4.43.1',
    'ImplantAssemblyTemplateStorage' : '1.2.840.10008.5.1.4.44.1',
    'ImplantTemplateGroupStorage' : '1.2.840.10008.5.1.4.45.1',
    'CTDefinedProcedureProtocolStorage' : '1.2.840.10008.5.1.4.1.1.200.1',
    'ProtocolApprovalStorage' : '1.2.840.10008.5.1.4.1.1.200.3',
}

_HANGING_PROTOCOL_CLASSES = {
    'HangingProtocolInformationModelFind' : '1.2.840.10008.5.1.4.38.2',
    'HangingProtocolInformationModelMove' : '1.2.840.10008.5.1.4.38.3',
    'HangingProtocolInformationModelGet' : '1.2.840.10008.5.1.4.38.4',
}

_DEFINED_PROCEDURE_CLASSES = {
    'DefinedProcedureProtocolInformationModelFind' : '1.2.840.10008.5.1.4.20.1',
    'DefinedProcedureProtocolInformationModelMove' : '1.2.840.10008.5.1.4.20.2',
    'DefinedProcedureProtocolInformationModelGet' : '1.2.840.10008.5.1.4.20.3',
}

_COLOR_PALETTE_CLASSES = {
    'ColorPaletteInformationModelFind' : '1.2.840.10008.5.1.4.39.2',
    'ColorPaletteInformationModelMove' : '1.2.840.10008.5.1.4.39.3',
    'ColorPaletteInformationModelGet' : '1.2.840.10008.5.1.4.39.4',
}

_IMPLANT_TEMPLATE_CLASSES = {
    'GenericImplantTemplateInformationModelFind' : '1.2.840.10008.5.1.4.43.2',
    'GenericImplantTemplateInformationModelMove' : '1.2.840.10008.5.1.4.43.3',
    'GenericImplantTemplateInformationModelGet' : '1.2.840.10008.5.1.4.43.4',
    'ImplantAssemblyTemplateInformationModelFind' : '1.2.840.10008.5.1.4.44.2',
    'ImplantAssemblyTemplateInformationModelMove' : '1.2.840.10008.5.1.4.44.3',
    'ImplantAssemblyTemplateInformationModelGet' : '1.2.840.10008.5.1.4.44.4',
    'ImplantTemplateGroupInformationModelFind' : '1.2.840.10008.5.1.4.45.2',
    'ImplantTemplateGroupInformationModelMove' : '1.2.840.10008.5.1.4.45.3',
    'ImplantTemplateGroupInformationModelGet' : '1.2.840.10008.5.1.4.45.4',
}

_DISPLAY_SYSTEM_CLASSES = {
    'DisplaySystemSOPClass' : '1.2.840.10008.5.1.1.40',
}

_PRINT_MANAGEMENT_CLASSES = {
    'BasicFilmSessionSOPClass' : '1.2.840.10008.5.1.1.1',
    'BasicFilmBoxSOPClass' : '1.2.840.10008.5.1.1.2',
    'BasicGrayscaleImageBoxSOPClass' : '1.2.840.10008.5.1.1.4',
    'BasicColourImageBoxSOPClass' : '1.2.840.10008.5.1.1.4.1',
    'PrintJobSOPClass' : '1.2.840.10008.5.1.1.14',
    'BasicAnnotationBoxSOPClass' : '1.2.840.10008.5.1.1.15',
    'PrinterSOPClass' : '1.2.840.10008.5.1.1.16',
    'PrinterConfigurationRetrievalSOPClass' : '1.2.840.10008.5.1.1.16.376',
    'PresentationLUTSOPClass' : '1.2.840.10008.5.1.1.23',
}

_PROCEDURE_STEP_CLASSES = {
    'ModalityPerformedProcedureStepSOPClass' : '1.2.840.10008.3.1.2.3.3',
    'ModalityPerformedProcedureStepRetrieveSOPClass' : '1.2.840.10008.3.1.2.3.4',
    'ModalityPerformedProcedureStepNotificationSOPClass' : '1.2.840.10008.3.1.2.3.5',
}

_MEDIA_STORAGE_CLASSES = {
    'MediaStorageDirectoryStorage' : '1.2.840.10008.1.3.10',
}

_UNIFIED_PROCEDURE_STEP_CLASSES = {
    'UnitedProcedureStepPush' : '1.2.840.10008.5.1.4.34.6.1',
    'UnitedProcedureStepWatch' : '1.2.840.10008.5.1.4.34.6.2',
    'UnitedProcedureStepPull' : '1.2.840.10008.5.1.4.34.6.3',
    'UnitedProcedureStepEvent' : '1.2.840.10008.5.1.4.34.6.4',
}

_RT_MACHINE_VERIFICATION_CLASSES = {
    'RTConventionalMachineVerification' : '1.2.840.10008.5.1.4.34.8',
    'RTIonMachineVerification' : '1.2.840.10008.5.1.4.34.9',
}

# pylint: enable=line-too-long
_generate_sop_classes(_VERIFICATION_CLASSES)
_generate_sop_classes(_STORAGE_CLASSES)
_generate_sop_classes(_QR_CLASSES)
_generate_sop_classes(_BASIC_WORKLIST_CLASSES)
_generate_sop_classes(_RELEVANT_PATIENT_QUERY_CLASSES)
_generate_sop_classes(_SUBSTANCE_ADMINISTRATION_CLASSES)
_generate_sop_classes(_NON_PATIENT_OBJECT_CLASSES)
_generate_sop_classes(_HANGING_PROTOCOL_CLASSES)
_generate_sop_classes(_DEFINED_PROCEDURE_CLASSES)
_generate_sop_classes(_COLOR_PALETTE_CLASSES)
_generate_sop_classes(_IMPLANT_TEMPLATE_CLASSES)
_generate_sop_classes(_DISPLAY_SYSTEM_CLASSES)
_generate_sop_classes(_PRINT_MANAGEMENT_CLASSES)
_generate_sop_classes(_PROCEDURE_STEP_CLASSES)
_generate_sop_classes(_MEDIA_STORAGE_CLASSES)
_generate_sop_classes(_UNIFIED_PROCEDURE_STEP_CLASSES)
_generate_sop_classes(_RT_MACHINE_VERIFICATION_CLASSES)


def uid_to_sop_class(uid):
    """Return the SOPClass object corresponding to `uid`.

    Parameters
    ----------
    uid : pydicom.uid.UID

    Returns
    -------
    sop_class.SOPClass subclass
        The SOP class corresponding to `uid`.

    Raises
    ------
    NotImplementedError
        If the SOP Class corresponding to the given UID has not been
        implemented.
    """
    # Get a list of all the class members of the current module
    members = inspect.getmembers(
        sys.modules[__name__],
        lambda mbr: isinstance(mbr, str)
    )

    for obj in members:
        if hasattr(obj[1], 'service_class') and obj[1] == uid:
            return obj[1]

    sop_class = SOPClass(uid)
    sop_class._service_class = ServiceClass

    return sop_class



"""Unit tests for the DIMSE Message classes."""

from io import BytesIO
import logging

import pytest

from pydicom.dataset import Dataset
from pydicom.uid import UID

from pynetdicom.dimse_messages import (
    C_STORE_RQ, C_STORE_RSP, DIMSEMessage, C_ECHO_RQ, C_ECHO_RSP, C_FIND_RQ,
    C_FIND_RSP, C_MOVE_RQ, C_MOVE_RSP, C_GET_RQ, C_GET_RSP, N_EVENT_REPORT_RQ,
    N_EVENT_REPORT_RSP, N_SET_RQ, N_SET_RSP, N_GET_RQ, N_GET_RSP, N_ACTION_RQ,
    N_ACTION_RSP, N_CREATE_RQ, N_CREATE_RSP, N_DELETE_RQ, N_DELETE_RSP,
    C_CANCEL_RQ
)
from pynetdicom.dimse_primitives import (
    C_STORE, C_ECHO, C_GET, C_MOVE, C_FIND, N_EVENT_REPORT, N_GET, N_SET,
    N_ACTION, N_CREATE, N_DELETE, C_CANCEL
)
from pynetdicom.dsutils import encode, decode
from pynetdicom.pdu_primitives import P_DATA
from pynetdicom.utils import pretty_bytes
from .encoded_dimse_msg import (
    c_echo_rq_cmd, c_echo_rsp_cmd, c_store_rq_cmd, c_store_ds, c_store_rsp_cmd,
    c_find_rq_cmd, c_find_rq_ds, c_find_rsp_cmd, c_find_rsp_ds, c_get_rq_cmd,
    c_get_rq_ds, c_get_rsp_cmd, c_get_rsp_ds, c_move_rq_cmd, c_move_rq_ds,
    c_move_rsp_cmd, c_move_rsp_ds
)
from .encoded_dimse_n_msg import (
    n_er_rq_cmd, n_er_rq_ds, n_er_rsp_cmd, n_er_rsp_ds,
    n_get_rq_cmd, n_get_rsp_cmd, n_get_rsp_ds,
    n_delete_rq_cmd, n_delete_rsp_cmd,
    n_action_rq_cmd, n_action_rq_ds, n_action_rsp_cmd, n_action_rsp_ds,
    n_create_rq_cmd, n_create_rq_ds, n_create_rsp_cmd, n_create_rsp_ds,
    n_set_rq_cmd, n_set_rq_ds, n_set_rsp_cmd, n_set_rsp_ds
)


LOGGER = logging.getLogger('pynetdicom')
LOGGER.setLevel(logging.CRITICAL)


def print_nice_bytes(bytestream):
    """Nice output for bytestream."""
    str_list = pretty_bytes(bytestream, prefix="b'\\x", delimiter='\\x',
                        items_per_line=10, suffix="' \\")
    for string in str_list:
        print(string)


class TestDIMSEMessage(object):
    """Test DIMSEMessage class"""
    def test_fragment_pdv(self):
        """Test that the PDV fragmenter is working correctly."""
        dimse_msg = C_STORE_RQ()
        frag = dimse_msg._generate_pdv_fragments

        result = []
        for fragment in frag(c_echo_rsp_cmd, 1000):
            result.append(fragment)
        assert len(result) == 1
        assert result[0] == c_echo_rsp_cmd
        assert isinstance(result[0], bytes)
        assert result[-1] != b''

        result = []
        for fragment in frag(c_echo_rsp_cmd, 10):
            result.append(fragment)
        assert len(result) == 20
        assert result[0] == c_echo_rsp_cmd[:4]
        assert isinstance(result[0], bytes)
        assert result[-1] != b''

        with pytest.raises(ValueError):
            next(frag(c_echo_rsp_cmd, 6))

    def test_fragment_pdv_zero(self):
        """Test that the PDV fragmenter works correctly for 0 max PDU."""
        dimse_msg = C_STORE_RQ()
        frag = dimse_msg._generate_pdv_fragments
        result = []
        for fragment in frag(c_echo_rsp_cmd, 0):
            result.append(fragment)

        assert len(result) == 1
        assert result[0] == c_echo_rsp_cmd

    def test_encode(self):
        """Test encoding of a DIMSE message."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = '1.1.1'
        primitive.AffectedSOPInstanceUID = '1.2.1'
        primitive.Priority = 0x02
        primitive.MoveOriginatorApplicationEntityTitle = 'UNITTEST'
        primitive.MoveOriginatorMessageID = 3

        # Test encode without dataset
        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(12, 16):
            p_data_list.append(pdata)
        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b'\x01'
        assert p_data_list[-1].presentation_data_value_list[0][1][0:1] == b'\x03'
        assert dimse_msg.context_id == 12

        # Test encode with dataset
        ds = Dataset()
        ds.PatientID = 'Test1101'
        ds.PatientName = 'Tube^HeNe'
        primitive.DataSet = BytesIO(encode(ds, True, True))

        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(13, 10):
            p_data_list.append(pdata)

        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b'\x01'
        assert p_data_list[-1].presentation_data_value_list[0][1][0:1] == b'\x02'
        assert p_data_list[-2].presentation_data_value_list[0][1][0:1] == b'\x00'
        assert p_data_list[-10].presentation_data_value_list[0][1][0:1] == b'\x03'
        assert dimse_msg.context_id == 13

        p_data_list = []
        for pdata in dimse_msg.encode_msg(1, 31682):
            p_data_list.append(pdata)
        assert p_data_list[0].presentation_data_value_list[0][1] == c_store_rq_cmd
        assert p_data_list[1].presentation_data_value_list[0][1] == c_store_ds

    def test_encode_zero(self):
        """Test encoding with a 0 max pdu length."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = '1.1.1'
        primitive.AffectedSOPInstanceUID = '1.2.1'
        primitive.Priority = 0x02
        primitive.MoveOriginatorApplicationEntityTitle = 'UNITTEST'
        primitive.MoveOriginatorMessageID = 3

        # Test encode without dataset
        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(12, 0):
            p_data_list.append(pdata)
        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b'\x03'
        assert dimse_msg.context_id == 12

        # Test encode with dataset
        ds = Dataset()
        ds.PatientID = 'Test1101'
        ds.PatientName = 'Tube^HeNe'
        primitive.DataSet = BytesIO(encode(ds, True, True))

        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)
        p_data_list = []
        for pdata in dimse_msg.encode_msg(13, 0):
            p_data_list.append(pdata)

        assert p_data_list[0].presentation_data_value_list[0][1][0:1] == b'\x03'
        assert p_data_list[0].presentation_data_value_list[0][1] == c_store_rq_cmd
        assert p_data_list[1].presentation_data_value_list[0][1][0:1] == b'\x02'
        assert p_data_list[1].presentation_data_value_list[0][1] == c_store_ds
        assert dimse_msg.context_id == 13

    def test_decode(self):
        """Test decoding of a DIMSE message."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = '1.1.1'
        primitive.AffectedSOPInstanceUID = '1.2.1'
        primitive.Priority = 0x02
        primitive.MoveOriginatorApplicationEntityTitle = b'UNITTEST'
        primitive.MoveOriginatorMessageID = 3
        ds = Dataset()
        ds.PatientID = 'Test1101'
        ds.PatientName = 'Tube^HeNe'
        primitive.DataSet = BytesIO(encode(ds, True, True))
        dimse_msg = C_STORE_RQ()
        dimse_msg.primitive_to_message(primitive)

        # CMD: (1x4, 3x1), DS: (0x1, 2x1)
        p_data = dimse_msg.encode_msg(12, 24)

        # Command set decoding
        dimse_msg.decode_msg(next(p_data)) # MCHB 1
        dimse_msg.decode_msg(next(p_data)) # MCHB 1
        dimse_msg.decode_msg(next(p_data)) # MCHB 1
        dimse_msg.decode_msg(next(p_data)) # MCHB 1
        dimse_msg.decode_msg(next(p_data)) # MCHB 3 - end of command set
        assert dimse_msg.__class__ == C_STORE_RQ

        # Test decoded command set
        cs = dimse_msg.command_set
        assert cs.CommandGroupLength == 102
        assert cs.AffectedSOPClassUID == UID('1.1.1')
        assert cs.AffectedSOPInstanceUID == UID('1.2.1')
        assert cs.Priority == 2
        assert cs.CommandDataSetType == 1
        assert cs.CommandField == 1
        assert cs.MoveOriginatorApplicationEntityTitle == 'UNITTEST        '
        assert cs.MoveOriginatorMessageID == 3

        # Test decoded dataset
        dimse_msg.decode_msg(next(p_data)) # MCHB 1 # MCHB 0
        dimse_msg.decode_msg(next(p_data)) # MCHB 1 # MCHB 2
        assert dimse_msg.data_set.getvalue() == c_store_ds[1:]

        # Test returns false
        msg = C_STORE_RSP()
        assert not msg.decode_msg(c_store_rsp_cmd)

    def test_primitive_to_message(self):
        """Test converting a DIMSE primitive to a DIMSE message."""
        primitive = C_STORE()
        primitive.MessageID = 7
        primitive.AffectedSOPClassUID = '1.1.1'
        primitive.AffectedSOPInstanceUID = '1.2.1'
        primitive.Priority = 0x02
        msg = C_STORE_RQ()
        msg.primitive_to_message(primitive)
        # Test unused command set elements are deleted
        assert not 'MoveOriginatorApplicationEntityTitle' in msg.command_set

        # Test raise error for unknown DIMSE message type
        msg.__class__.__name__ = 'TestClass'
        with pytest.raises(ValueError):
            msg.primitive_to_message(primitive)

        # Reset name to avoid errors in other tests
        msg.__class__.__name__ = 'C_STORE_RQ'

    # DIMSE-C
    def test_message_to_primitive_c_store(self):
        """Test converting C_STORE_RQ and _RSP to C_STORE primitive."""
        msg = C_STORE_RQ()
        for data in [c_store_rq_cmd, c_store_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_STORE)
        assert isinstance(primitive.DataSet, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.1.1')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.1')
        assert primitive.Priority == 2
        assert primitive.MoveOriginatorApplicationEntityTitle == b'UNITTEST        '
        assert primitive.MoveOriginatorMessageID == 3

        ds = decode(primitive.DataSet, True, True)
        assert ds.PatientName == 'Tube^HeNe'
        assert ds.PatientID == 'Test1101'

        msg = C_STORE_RSP()
        p_data = P_DATA()
        p_data.presentation_data_value_list.append([0, c_store_rsp_cmd])
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_STORE)
        assert primitive.DataSet is None
        for elem in msg.command_set:
            if hasattr(primitive, elem.keyword):
                item = getattr(primitive, elem.keyword)
                assert item == elem.value

    def test_message_to_primitive_c_echo(self):
        """Test converting C_ECHO_RQ to C_ECHO primitive."""
        msg = C_ECHO_RQ()
        p_data = P_DATA()
        p_data.presentation_data_value_list.append([0, c_echo_rq_cmd])
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_ECHO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.1.1')
        assert primitive.MessageID == 7

        msg = C_ECHO_RSP()
        p_data = P_DATA()
        p_data.presentation_data_value_list.append([0, c_echo_rsp_cmd])
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_ECHO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.1.1')
        assert primitive.MessageIDBeingRespondedTo == 8
        assert primitive.Status == 0

    def test_message_to_primitive_c_get(self):
        """Test converting C_GET_RQ to C_GET primitive."""
        msg = C_GET_RQ()
        for data in [c_get_rq_cmd, c_get_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_GET)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.Priority == 2
        assert primitive.MessageID == 7

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == 'PATIENT'
        assert ds.PatientID == '*'

        msg = C_GET_RSP()
        for data in [c_get_rsp_cmd, c_get_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_GET)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.Status == 65280
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.NumberOfRemainingSuboperations == 3
        assert primitive.NumberOfCompletedSuboperations == 1
        assert primitive.NumberOfFailedSuboperations == 2
        assert primitive.NumberOfWarningSuboperations == 4

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == 'PATIENT'
        assert ds.PatientID == '*'

    def test_message_to_primitive_c_move(self):
        """Test converting C_MOVE_RQ to C_MOVE primitive."""
        msg = C_MOVE_RQ()
        for data in [c_move_rq_cmd, c_move_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_MOVE)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.Priority == 2
        assert primitive.MoveDestination == b'MOVE_SCP        '
        assert primitive.MessageID == 7

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == 'PATIENT'
        assert ds.PatientID == '*'

        msg = C_MOVE_RSP()
        for data in [c_move_rsp_cmd, c_move_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_MOVE)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.Status == 65280
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.NumberOfRemainingSuboperations == 3
        assert primitive.NumberOfCompletedSuboperations == 1
        assert primitive.NumberOfFailedSuboperations == 2
        assert primitive.NumberOfWarningSuboperations == 4

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == 'PATIENT'
        assert ds.PatientID == '*'

    def test_message_to_primitive_c_find(self):
        """Test converting C_FIND_RQ to C_FIND primitive."""
        msg = C_FIND_RQ()
        for data in [c_find_rq_cmd, c_find_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_FIND)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.Priority == 2
        assert primitive.MessageID == 7

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == 'PATIENT'
        assert ds.PatientID == '*'

        msg = C_GET_RSP()
        for data in [c_find_rsp_cmd, c_find_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_FIND)
        assert isinstance(primitive.Identifier, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.Status == 65280

        ds = decode(primitive.Identifier, True, True)
        assert ds.QueryRetrieveLevel == 'PATIENT'
        assert ds.RetrieveAETitle == 'FINDSCP'

        assert ds.PatientName == 'ANON^A^B^C^D'

    def test_message_to_primitive_c_cancel(self):
        """Test converting C_CANCEL_rq to C_CANCEL primitive."""
        msg = C_CANCEL_RQ()
        cs = Dataset()
        cs.MessageIDBeingRespondedTo = 1
        msg.command_set = cs
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, C_CANCEL)
        assert primitive.MessageIDBeingRespondedTo == 1

    # DIMSE-N
    def test_message_to_primitive_n_event_report(self):
        """Test converting N_EVENT_REPORT_RQ and _RSP to primitive."""
        # N-EVENT-REPORT-RQ
        msg = N_EVENT_REPORT_RQ()
        for data in [n_er_rq_cmd, n_er_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_EVENT_REPORT)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.392.200036.9116.2.6.1.48')
        assert primitive.MessageID == 7
        assert primitive.EventTypeID == 2

        ds = decode(primitive.EventInformation, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

        # N-EVENT-REPORT-RSP
        msg = N_EVENT_REPORT_RSP()
        for data in [n_er_rsp_cmd, n_er_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_EVENT_REPORT)
        assert primitive.AffectedSOPClassUID == UID('1.2.4.10')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.4.5.7.8')
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.EventTypeID == 2
        assert primitive.Status == 0x0000

        ds = decode(primitive.EventReply, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

    def test_message_to_primitive_n_get(self):
        """Test converting N_GET_RQ and _RSP to primitive."""
        # N-GET-RQ
        msg = N_GET_RQ()
        for data in [n_get_rq_cmd]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_GET)
        assert primitive.RequestedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.RequestedSOPInstanceUID == UID('1.2.392.200036.9116.2.6.1.48')
        assert primitive.MessageID == 7
        primitive.AttributeIdentifierList = [
            (0x7fe0,0x0010), (0x0000,0x0000), (0xFFFF,0xFFFF)
        ]

        msg = N_GET_RSP()
        for data in [n_get_rsp_cmd, n_get_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_GET)
        assert primitive.AttributeIdentifierList is None
        assert primitive.AffectedSOPClassUID == UID('1.2.4.10')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.4.5.7.8')
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0x0000

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

    def test_message_to_primitive_n_set(self):
        """Test converting N_SET_RQ and _RSP to primitive."""
        # N-SET-RQ
        msg = N_SET_RQ()
        for data in [n_set_rq_cmd, n_set_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_SET)
        assert primitive.RequestedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.RequestedSOPInstanceUID == UID('1.2.392.200036.9116.2.6.1.48')
        assert primitive.MessageID == 7

        ds = decode(primitive.ModificationList, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

        # N-SET-RSP
        msg = N_SET_RSP()
        for data in [n_set_rsp_cmd, n_set_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_SET)
        assert isinstance(primitive.AttributeList, BytesIO)
        assert primitive.AffectedSOPClassUID == UID('1.2.4.10')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.4.5.7.8')
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0x0000

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

    def test_message_to_primitive_n_action(self):
        """Test converting N_ACTION_RQ and _RSP to primitive."""
        # N-ACTION-RQ
        msg = N_ACTION_RQ()
        for data in [n_action_rq_cmd, n_action_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_ACTION)
        assert primitive.RequestedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.RequestedSOPInstanceUID == UID('1.2.392.200036.9116.2.6.1.48')
        assert primitive.MessageID == 7
        assert primitive.ActionTypeID == 1

        ds = decode(primitive.ActionInformation, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

        # N-ACTION-RSP
        msg = N_ACTION_RSP()
        for data in [n_action_rsp_cmd, n_action_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_ACTION)
        assert primitive.AffectedSOPClassUID == UID('1.2.4.10')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.4.5.7.8')
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.ActionTypeID == 1
        assert primitive.Status == 0x0000

        ds = decode(primitive.ActionReply, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

    def test_message_to_primitive_n_create(self):
        """Test converting N_CREATE_RQ and _RSP to primitive."""
        # N-CREATE-RQ
        msg = N_CREATE_RQ()
        for data in [n_create_rq_cmd, n_create_rq_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_CREATE)
        assert primitive.AffectedSOPClassUID == UID('1.2.840.10008.5.1.4.1.1.2')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.392.200036.9116.2.6.1.48')
        assert primitive.MessageID == 7

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

        # N-ACTION-RSP
        msg = N_CREATE_RSP()
        for data in [n_create_rsp_cmd, n_create_rsp_ds]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_CREATE)
        assert primitive.AffectedSOPClassUID == UID('1.2.4.10')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.4.5.7.8')
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0x0000

        ds = decode(primitive.AttributeList, True, True)
        assert ds.PatientName == 'Tube HeNe'
        assert ds.PatientID == 'Test1101'

    def test_message_to_primitive_n_delete(self):
        """Test converting N_DELETE_RQ and _RSP to primitive."""
        # N-DELETE-RQ
        msg = N_DELETE_RQ()
        for data in [n_delete_rq_cmd]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_DELETE)
        assert primitive.RequestedSOPClassUID == UID('1.2.3')
        assert primitive.RequestedSOPInstanceUID == UID('1.2.30')
        assert primitive.MessageID == 7

        # N-DELETE-RSP
        msg = N_DELETE_RSP()
        for data in [n_delete_rsp_cmd]:
            p_data = P_DATA()
            p_data.presentation_data_value_list.append([0, data])
            msg.decode_msg(p_data)
        primitive = msg.message_to_primitive()
        assert isinstance(primitive, N_DELETE)
        assert primitive.AffectedSOPClassUID == UID('1.2.4.10')
        assert primitive.AffectedSOPInstanceUID == UID('1.2.4.5.7.8')
        assert primitive.MessageIDBeingRespondedTo == 5
        assert primitive.Status == 0xC201


class TestThreadSafety(object):
    """Tests for the thread safety of DIMSEMessage classes."""
    def setup(self):
        ds = Dataset()
        ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1'
        ds.SOPInstanceUID = '1.2.3.4'
        ds.PatientName = 'Test^Testing'
        self.ds = ds

    def test_msg_init(self):
        """Test the DIMSE Message dynamic construction works OK."""
        msg = C_STORE_RQ()
        cs = msg.command_set
        ds = msg.data_set
        assert cs.CommandGroupLength is None
        assert cs.AffectedSOPClassUID == ''
        assert cs.CommandField is None
        assert cs.MessageID is None
        assert cs.Priority is None
        assert cs.CommandDataSetType is None
        assert cs.AffectedSOPInstanceUID == ''
        assert cs.MoveOriginatorApplicationEntityTitle is None
        assert cs.MoveOriginatorMessageID is None
        assert ds.getvalue() == b''

    def test_cstore_rq(self):
        """Test creation of a DIMSE Message from a primitive works OK."""
        # Testing that primitive_to_message doesn't affect the original class
        #   definition
        req = C_STORE()
        req.MessageID = 12
        req.AffectedSOPClassUID = self.ds.SOPClassUID
        req.AffectedSOPInstanceUID = self.ds.SOPInstanceUID
        req.Priority = 0x02
        req.MoveOriginatorApplicationEntityTitle = None
        req.MoveOriginatorMessageID = None
        req.DataSet = BytesIO(encode(self.ds, True, True))

        msg = C_STORE_RQ()
        msg.primitive_to_message(req)
        cs = msg.command_set
        assert cs.CommandGroupLength == 88
        assert cs.AffectedSOPClassUID == '1.2.840.10008.5.1.4.1.1'
        assert cs.CommandField == 1
        assert cs.MessageID == 12
        assert cs.Priority == 2
        assert cs.CommandDataSetType == 1
        assert cs.AffectedSOPInstanceUID == '1.2.3.4'
        assert 'MoveOriginatorApplicationEntityTitle' not in cs
        assert 'MoveOriginatorMessageID' not in cs
        assert msg.data_set.getvalue() != b''

        # Test new class instance is OK
        msg = C_STORE_RQ()
        cs = msg.command_set
        ds = msg.data_set
        assert cs.CommandGroupLength is None
        assert cs.AffectedSOPClassUID == ''
        assert cs.CommandField is None
        assert cs.MessageID is None
        assert cs.Priority is None
        assert cs.CommandDataSetType is None
        assert cs.AffectedSOPInstanceUID == ''
        assert cs.MoveOriginatorApplicationEntityTitle is None
        assert cs.MoveOriginatorMessageID is None
        assert ds.getvalue() == b''

    def test_message_builder_regression(self):
        """Regression test for DIMSEMessage class builder."""
        with pytest.raises(AttributeError, match=r"no attribute 'command_set'"):
            assert C_STORE_RQ.command_set.MessageID
        with pytest.raises(AttributeError, match=r"no attribute 'data_set'"):
            assert C_STORE_RQ.data_set.get_value() == b''



"""
Implementation of the DIMSE service provider.
"""
from io import BytesIO
import logging
try:
    import queue
except ImportError:
    import Queue as queue  # Python 2 compatibility
import time

# pylint: disable=no-name-in-module
# TODO: Consider switching to * import, check that _var aren't shown (__var?)
from pynetdicom.dimse_messages import (
    C_STORE_RQ, C_STORE_RSP, C_FIND_RQ, C_FIND_RSP, C_GET_RQ, C_GET_RSP,
    C_MOVE_RQ, C_MOVE_RSP, C_ECHO_RQ, C_ECHO_RSP, C_CANCEL_RQ,
    N_EVENT_REPORT_RQ, N_EVENT_REPORT_RSP, N_GET_RQ, N_GET_RSP, N_SET_RQ,
    N_SET_RSP, N_ACTION_RQ, N_ACTION_RSP, N_CREATE_RQ, N_CREATE_RSP,
    N_DELETE_RQ, N_DELETE_RSP, DIMSEMessage
)
# pylint: enable=no-name-in-module
from pynetdicom.dimse_primitives import (
    C_STORE, C_FIND, C_GET, C_MOVE, C_ECHO, N_EVENT_REPORT, N_GET, N_SET,
    N_ACTION, N_CREATE, N_DELETE, C_CANCEL
)
from pynetdicom.pdu_primitives import P_DATA
from pynetdicom.sop_class import uid_to_service_class
from pynetdicom.timer import Timer
from pynetdicom._globals import DEFAULT_MAX_LENGTH

LOGGER = logging.getLogger('pynetdicom.dimse')

_RQ_TO_MESSAGE = {
    C_ECHO : C_ECHO_RQ,
    C_STORE : C_STORE_RQ,
    C_FIND : C_FIND_RQ,
    C_MOVE : C_MOVE_RQ,
    C_GET : C_GET_RQ,
    N_EVENT_REPORT : N_EVENT_REPORT_RQ,
    N_GET : N_GET_RQ,
    N_SET : N_SET_RQ,
    N_ACTION : N_ACTION_RQ,
    N_CREATE : N_CREATE_RQ,
    N_DELETE : N_DELETE_RQ
}
_RSP_TO_MESSAGE = {
    C_ECHO : C_ECHO_RSP,
    C_STORE : C_STORE_RSP,
    C_FIND : C_FIND_RSP,
    C_MOVE : C_MOVE_RSP,
    C_GET : C_GET_RSP,
    C_CANCEL : C_CANCEL_RQ,
    N_EVENT_REPORT : N_EVENT_REPORT_RSP,
    N_GET : N_GET_RSP,
    N_SET : N_SET_RSP,
    N_ACTION : N_ACTION_RSP,
    N_CREATE : N_CREATE_RSP,
    N_DELETE : N_DELETE_RSP
}


class DIMSEServiceProvider(object):
    """The DIMSE service provider.

    A DICOM AE uses the services provided by the DICOM Message Service Element
    (DIMSE). DIMSE specifies two sets of services.

    - DIMSE-C supports operations associated with composite SOP Classes and
      provides effective compatibility with the previous versions of the DICOM
      standards.
    - DIMSE-N supports operations associated with normalised SOP Classes and
      provides an extended set of object-orientated operations and
      notifications

    **Service Overview**

    The DIMSE service provider supports communication between peer DIMSE
    service users. A service user acts in one of two roles:

    - invoking DIMSE user
    - performing DIMSE user

    Service users make use of service primitives provided by the DIMSE service
    provider. A service primitive shall be one of the following types:

    - request primitive
    - indication primitive
    - response primitive
    - confirmation primitive

    These primitives are used as follows:

    - The invoking service user issues a request primitive to the DIMSE
      provider
    - The DIMSE provider receives the request primitive and issues an
      indication primitive to the performing service user
    - The performing service user receives the indication primitive and
      performs the requested service
    - The performing service user issues a response primitive to the DIMSE
      provider
    - The DIMSE provider receives the response primitive and issues a
      confirmation primitive to the invoking service user
    - The invoking service user receives the confirmation primitive, completing
      the DIMSE service.

    **Service Primitive Classes**

    * DIMSE-C: C_ECHO, C_STORE, C_GET, C_FIND, C_MOVE
    * DIMSE-N: N_EVENT_REPORT, N_GET, N_GET, N_ACTION, N_CREATE, N_DELETE

    **Protocol Machine**

    PS3.7 8.1
    The DIMSE protocol machine defines the procedures and the encoding rules
    necessary to construct Messages used to exchange command requests and
    responses between peer DIMSE service users.

    The DIMSE protocol machine accepts service user requests and response
    service primitives and constructs Messages defined by the procedures in
    PS3.7 9.3 and 10.3. The DIMSE protocol machine accepts Messages and passes
    them to the DIMSE service user by the means of indication and confirmation
    service primitives.

    **Messages**

    +----------------+-----------------------+------------------------+
    | Primitive      | Type                  | Message Class          |
    +----------------+-----------------------+------------------------+
    | C-CANCEL       | Request/indication    | C_CANCEL_R             |
    +----------------+-----------------------+------------------------+
    | C-ECHO         | Request/indication    | C_ECHO_RQ              |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_ECHO_RSP             |
    +----------------+-----------------------+------------------------+
    | C-FIND         | Request/indication    | C_FIND_RQ              |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_FIND_RSP             |
    +----------------+-----------------------+------------------------+
    | C-GET          | Request/indication    | C_GET_RQ               |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_GET_RSP              |
    +----------------+-----------------------+------------------------+
    | C-MOVE         | Request/indication    | C_MOVE_RQ              |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_MOVE_RSP             |
    +----------------+-----------------------+------------------------+
    | C-STORE        | Request/indication    | C_STORE_RQ             |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_STORE_RSP            |
    +----------------+-----------------------+------------------------+
    | N-ACTION       | Request/indication    | N_ACTION_RQ            |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_ACTION_RSP           |
    +----------------+-----------------------+------------------------+
    | N-CREATE       | Request/indication    | N_CREATE_RQ            |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_CREATE_RSP           |
    +----------------+-----------------------+------------------------+
    | N-DELETE       | Request/indication    | N_DELETE_RQ            |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_DELETE_RSP           |
    +----------------+-----------------------+------------------------+
    | N-EVENT-REPORT | Request/indication    | N_EVENT_REPORT_RQ      |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_EVENT_REPORT_RSP     |
    +----------------+-----------------------+------------------------+
    | N-GET          | Request/indication    | N_GET_RQ               |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_GET_RSP              |
    +----------------+-----------------------+------------------------+
    | N-SET          | Request/indication    | N_SET_RQ               |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_SET_RSP              |
    +----------------+-----------------------+------------------------+

    Attributes
    ----------
    cancel_rq : dict
        A dict of {MessageIDBeingRespondedTo : C_CANCEL} messages received.
        The dict is cleared out at the start and end of Service Class
        operations and is limited to a maximum of 10 messages.
    dimse_timeout : numeric or None
        The number of seconds before the DIMSE service timeout. A value of
        ``None`` indicates no timeout.
    dul : dul.DULServiceProvider
        The DICOM Upper Layer service provider.
    maximum_pdu_size : int
            The maximum PDU size when sending DIMSE messages.
    message : dimse_messages.DIMSEMessage
        The DIMSE message.
    msg_queue: queue.queue of dimse_messages.DIMSEMessage
        A queue holding decoded DIMSE Message primitives received from the
        peer, except for C-CANCEL requests.

    References
    ----------

    * DICOM Standard, Part 7
    """
    # pylint: disable=too-many-public-methods
    def __init__(self, dul, dimse_timeout=None,
                 maximum_pdu_size=DEFAULT_MAX_LENGTH):
        """Start the DIMSE service provider.

        Parameters
        ----------
        dul : dul.DULServiceProvider
            The DICOM Upper Layer service provider.
        dimse_timeout : int or float or None
            The number of seconds before the DIMSE service timeout. A value of
            None indicates no timeout.
        maximum_pdu_size : int
            The maximum PDU size when sending DIMSE messages, default 16382.
        """
        self.cancel_req = {}
        self.dimse_timeout = dimse_timeout
        self.dul = dul
        self.maximum_pdu_size = maximum_pdu_size
        self.message = None
        self.msg_queue = queue.Queue()

    def get_msg(self, block=False):
        """Get the next available DIMSE message.

        Parameters
        ----------
        block : bool
            If True then the function will block until either a message is
            available or `dimse_timeout` expires, otherwise non-blocking.

        Returns
        -------
        int, dimse_messages.DIMSEMessage or None, None
            The next available (context ID, DIMSE message), which is taken off
            the queue, or (None, None) if no messages are available within
            the `dimse_timeout` period.
        """
        try:
            return self.msg_queue.get(block=block, timeout=self.dimse_timeout)
        except queue.Empty:
            return None, None

    def peek_msg(self):
        """Return the first message in the message queue or None.

        Returns
        -------
        int, dimse_messages.DIMSEMessage or None, None
            The first (context ID, message) in the queue if one is available,
            otherwise (None, None). No messages are taken out of the queue.
        """
        try:
            return self.msg_queue.queue[0]
        except (queue.Empty, IndexError):
            return None, None

    def receive_primitive(self, primitive):
        """Process a P-DATA primitive received from the remote.

        A DIMSE message is split into one or more P-DATA primitives, which
        must be sent in sequential order. While waiting for all the P-DATA
        primitives associated with a message the encoded data is stored in
        self.message, which is decoded only when complete and converted
        into a DIMSE Message primitive which is added to the `msg_queue`.

        This makes it possible to process incoming P-DATA primitives into
        DIMSE messages while a service class implementation is running.

        Parameters
        ----------
        primitive : pdu_primitives.P_DATA
            A P-DATA primitive received from the peer to be processed.
        """
        if self.message is None:
            self.message = DIMSEMessage()

        if self.message.decode_msg(primitive):
            # Callback
            # TODO: Make this a package level option to increase speed
            self.on_receive_dimse_message(self.message)

            context_id = self.message.context_id
            try:
                primitive = self.message.message_to_primitive()
            except Exception as exc:
                LOGGER.error("Received an invalid DIMSE message")
                LOGGER.exception(exc)
                self.dul.event_queue.put('Evt19')
                return

            # Keep C-CANCEL requests separate from other messages
            # Only allow up to 10 C-CANCEL requests
            if isinstance(primitive, C_CANCEL) and len(self.cancel_req) < 10:
                msg_id = primitive.MessageIDBeingRespondedTo
                self.cancel_req[msg_id] = primitive
            else:
                self.msg_queue.put((context_id, primitive))

            # Fix for memory leak, Issue #41
            #   Reset the DIMSE message, ready for the next one
            self.message.encoded_command_set = BytesIO()
            self.message.data_set = BytesIO()
            self.message = None

    def send_msg(self, primitive, context_id):
        """Send a DIMSE-C or DIMSE-N message to the peer AE.

        Parameters
        ----------
        primitive : dimse_primitives DIMSE Primitive class
            The DIMSE message primitive to send to the peer.
        context_id : int
            The ID of the presentation context that the message is to be
            sent under.
        """
        if primitive.MessageIDBeingRespondedTo is None:
            dimse_msg = _RQ_TO_MESSAGE[primitive.__class__]()
        else:
            dimse_msg = _RSP_TO_MESSAGE[primitive.__class__]()

        # Convert DIMSE primitive to DIMSE Message
        dimse_msg.primitive_to_message(primitive)
        dimse_msg.context_id = context_id

        # Callbacks
        self.on_send_dimse_message(dimse_msg)

        # Split the full messages into P-DATA chunks,
        #   each below the max_pdu size
        for pdata in dimse_msg.encode_msg(context_id, self.maximum_pdu_size):
            self.dul.send_pdu(pdata)

    # Debugging and AE callbacks
    def on_send_dimse_message(self, message):
        """Controls which debugging function is called when sending.

        Will be called immediately prior to encoding and sending a DIMSE
        message.

        Parameters
        ----------
        message : dimse_messages.DIMSEMessage
            The DIMSE message to be sent.
        """
        callback = {C_ECHO_RQ: self.debug_send_c_echo_rq,
                    C_ECHO_RSP: self.debug_send_c_echo_rsp,
                    C_FIND_RQ: self.debug_send_c_find_rq,
                    C_FIND_RSP: self.debug_send_c_find_rsp,
                    C_GET_RQ: self.debug_send_c_get_rq,
                    C_GET_RSP: self.debug_send_c_get_rsp,
                    C_MOVE_RQ: self.debug_send_c_move_rq,
                    C_MOVE_RSP: self.debug_send_c_move_rsp,
                    C_STORE_RQ: self.debug_send_c_store_rq,
                    C_STORE_RSP: self.debug_send_c_store_rsp,
                    C_CANCEL_RQ: self.debug_send_c_cancel_rq,
                    N_EVENT_REPORT_RQ: self.debug_send_n_event_report_rq,
                    N_EVENT_REPORT_RSP: self.debug_send_n_event_report_rsp,
                    N_SET_RQ: self.debug_send_n_set_rq,
                    N_SET_RSP: self.debug_send_n_set_rsp,
                    N_GET_RQ: self.debug_send_n_get_rq,
                    N_GET_RSP: self.debug_send_n_get_rsp,
                    N_ACTION_RQ: self.debug_send_n_action_rq,
                    N_ACTION_RSP: self.debug_send_n_action_rsp,
                    N_CREATE_RQ: self.debug_send_n_create_rq,
                    N_CREATE_RSP: self.debug_send_n_create_rsp,
                    N_DELETE_RQ: self.debug_send_n_delete_rq,
                    N_DELETE_RSP: self.debug_send_n_delete_rsp}

        callback[type(message)](message)

    def on_receive_dimse_message(self, message):
        """Controls which debugging function is called when receiving.

        Function will be called immediately after receiving and decoding a
        DIMSE message.

        Parameters
        ----------
        message : dimse_messages.DIMSEMessage
            The DIMSE message that was received.
        """
        callback = {C_ECHO_RQ: self.debug_receive_c_echo_rq,
                    C_ECHO_RSP: self.debug_receive_c_echo_rsp,
                    C_FIND_RQ: self.debug_receive_c_find_rq,
                    C_FIND_RSP: self.debug_receive_c_find_rsp,
                    C_CANCEL_RQ: self.debug_receive_c_cancel_rq,
                    C_GET_RQ: self.debug_receive_c_get_rq,
                    C_GET_RSP: self.debug_receive_c_get_rsp,
                    C_MOVE_RQ: self.debug_receive_c_move_rq,
                    C_MOVE_RSP: self.debug_receive_c_move_rsp,
                    C_STORE_RQ: self.debug_receive_c_store_rq,
                    C_STORE_RSP: self.debug_receive_c_store_rsp,
                    N_EVENT_REPORT_RQ: self.debug_receive_n_event_report_rq,
                    N_EVENT_REPORT_RSP: self.debug_receive_n_event_report_rsp,
                    N_SET_RQ: self.debug_receive_n_set_rq,
                    N_SET_RSP: self.debug_receive_n_set_rsp,
                    N_GET_RQ: self.debug_receive_n_get_rq,
                    N_GET_RSP: self.debug_receive_n_get_rsp,
                    N_ACTION_RQ: self.debug_receive_n_action_rq,
                    N_ACTION_RSP: self.debug_receive_n_action_rsp,
                    N_CREATE_RQ: self.debug_receive_n_create_rq,
                    N_CREATE_RSP: self.debug_receive_n_create_rsp,
                    N_DELETE_RQ: self.debug_receive_n_delete_rq,
                    N_DELETE_RSP: self.debug_receive_n_delete_rsp}

        callback[type(message)](message)

    # Mid-level DIMSE Message related logging/debugging
    # pylint: disable=unused-argument
    @staticmethod
    def debug_send_c_echo_rq(msg):
        """Debugging function when a C-ECHO-RQ is sent.

        **C-ECHO Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID

        Parameters
        ----------
        msg : dimse_messages.C_ECHO_RQ
            The C-ECHO-RQ message to be sent.
        """
        cs = msg.command_set
        LOGGER.info("Sending Echo Request: MsgID %s", cs.MessageID)

    @staticmethod
    def debug_send_c_echo_rsp(msg):
        """Debugging function when a C-ECHO-RSP is sent.

        **C-ECHO Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (M) Status

        Parameters
        ----------
        msg : dimse_messages.C_ECHO_RSP
            The C-ECHO-RSP message to be sent.
        """
        pass

    @staticmethod
    def debug_send_c_store_rq(msg):
        """Debugging function when a C-STORE-RQ is sent.

        **C-STORE Request Elements**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Affected SOP Instance UID
        | (M) Priority
        | (U) Move Originator Application Entity Title
        | (U) Move Originator Message ID
        | (M) Data Set

        Parameters
        ----------
        msg : dimse_messages.C_STORE_RQ
            The C-STORE-RQ message to be sent.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        if cs.AffectedSOPClassUID.name == 'CT Image Storage':
            dataset_type = ', (CT)'
        elif cs.AffectedSOPClassUID.name == 'MR Image Storage':
            dataset_type = ', (MR)'
        else:
            dataset_type = ''

        LOGGER.info("Sending Store Request: MsgID %s%s",
                    cs.MessageID, dataset_type)

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-STORE RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Affected SOP Instance UID     : {0!s}'
                 .format(cs.AffectedSOPInstanceUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('Priority                      : {0!s}'.format(priority))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_c_store_rsp(msg):
        """Debugging function when a C-STORE-RSP is sent.

        **C-STORE Response Elements**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Affected SOP Instance UID
        | (M) Status

        Parameters
        ----------
        msg : dimse_messages.C_STORE_RSP
            The C-STORE-RSP message to be sent.
        """
        pass

    @staticmethod
    def debug_send_c_find_rq(msg):
        """Debugging function when a C-FIND-RQ is sent.

        **C-FIND Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Priority
        | (M) Identifier

        Parameters
        ----------
        msg : dimse_messages.C_FIND_RQ
            The C-FIND-RQ message to be sent.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        LOGGER.info("Sending Find Request: MsgID %s", cs.MessageID)

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-FIND RQ'))
        s.append('Presentation Context ID       : {0!s}'
                 .format(msg.context_id))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Identifier                    : {0!s}'.format(dataset))
        s.append('Priority                      : {0!s}'.format(priority))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_c_find_rsp(msg):
        """Debugging function when a C-FIND-RSP is sent.

        **C-FIND Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (C) Identifier
        | (M) Status

        Parameters
        ----------
        msg : dimse_messages.C_FIND_RSP
            The C-FIND-RSP message to be sent.
        """
        cs = msg.command_set

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-FIND RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('DIMSE Status                  : 0x{0:04x}'.format(cs.Status))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_c_get_rq(msg):
        """Debugging function when a C-GET-RQ is sent.

        **C-GET Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Priority
        | (M) Identifier

        Parameters
        ----------
        msg : dimse_messages.C_GET_RQ
            The C-GET-RQ message to be sent.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        LOGGER.info("Sending Get Request: MsgID %s", cs.MessageID)

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-GET RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('Priority                      : {0!s}'.format(priority))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_c_get_rsp(msg):
        """Debugging function when a C-GET-RSP is sent.

        **C-GET Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Identifier
        | (M) Status
        | (C) Number of Remaining Sub-operations
        | (C) Number of Completed Sub-operations
        | (C) Number of Failed Sub-operations
        | (C) Number of Warning Sub-operations

        Parameters
        ----------
        msg : dimse_messages.C_GET_RSP
            The C-GET-RSP message to be sent.
        """
        cs = msg.command_set

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        affected_sop = getattr(cs, 'AffectedSOPClassUID', 'None')

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-GET RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(affected_sop))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('DIMSE Status                  : 0x{0:04x}'.format(cs.Status))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_c_move_rq(msg):
        """Debugging function when a C-MOVE-RQ is sent.

        **C-MOVE Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Priority
        | (M) Move Destination
        | (M) Identifier

        Parameters
        ----------
        msg : dimse_messages.C_MOVE_RQ
            The C-MOVE-RQ message to be sent.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        identifier = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            identifier = 'Present'

        LOGGER.info("Sending Move Request: MsgID %s", cs.MessageID)

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-MOVE RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Move Destination              : {0!s}'
                 .format(cs.MoveDestination))
        s.append('Identifier                    : {0!s}'.format(identifier))
        s.append('Priority                      : {0!s}'.format(priority))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)
        return None

    @staticmethod
    def debug_send_c_move_rsp(msg):
        """Debugging function when a C-MOVE-RSP is sent.

        **C-MOVE Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Identifier
        | (M) Status
        | (C) Number of Remaining Sub-operations
        | (C) Number of Completed Sub-operations
        | (C) Number of Failed Sub-operations
        | (C) Number of Warning Sub-operations

        Parameters
        ----------
        msg : dimse_messages.C_MOVE_RSP
            The C-MOVE-RSP message to be sent.
        """
        cs = msg.command_set

        identifier = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            identifier = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-MOVE RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        else:
            s.append('Affected SOP Class UID        : none')
        s.append('Identifier                    : {0!s}'.format(identifier))
        s.append('Status                        : 0x{0:04x}'.format(cs.Status))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_c_cancel_rq(msg):
        """Debugging function when a C-CANCEL-RQ is sent.

        Covers C-CANCEL-FIND-RQ, C-CANCEL-GET-RQ and C-CANCEL-MOVE-RQ.

        Parameters
        ----------
        msg : dimse_messages.C_CANCEL_RQ
            The C-CANCEL-RQ message to be sent.
        """
        pass

    @staticmethod
    def debug_receive_c_echo_rq(msg):
        """Debugging function when a C-ECHO-RQ is received.

        **C-ECHO Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID

        Parameters
        ----------
        msg : dimse_messagesC_ECHO_RQ
            The received C-ECHO-RQ message.
        """
        cs = msg.command_set

        LOGGER.info('Received Echo Request (MsgID %s)', cs.MessageID)

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-ECHO RQ'))
        s.append('Presentation Context ID       : {0!s}'
                 .format(msg.context_id))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Data Set                      : {0!s}'.format('none'))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_receive_c_echo_rsp(msg):
        """Debugging function when a C-ECHO-RSP is received.

        **C-ECHO Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (M) Status

        Parameters
        ----------
        msg : dimse_messages.C_ECHO_RSP
            The received C-ECHO-RSP message.
        """
        cs = msg.command_set
        # Status is one of the following:
        #   0x0000 Success
        #   0x0122 Refused: SOP Class Not Supported
        #   0x0210 Refused: Duplicate Invocation
        #   0x0212 Refused: Mistyped Argument
        #   0x0211 Refused: Unrecognised Operation
        if cs.Status == 0x0000:
            status_str = 'Success'
        else:
            status_str = '0x{0:04x} - Unknown'.format(cs.Status)
        LOGGER.info("Received Echo Response (Status: %s)", status_str)

    @staticmethod
    def debug_receive_c_store_rq(msg):
        """Debugging function when a C-STORE-RQ is received.

        **C-STORE Request Elements**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Affected SOP Instance UID
        | (M) Priority
        | (U) Move Originator Application Entity Title
        | (U) Move Originator Message ID
        | (M) Data Set

        Parameters
        ----------
        msg : dimse_messagesC_STORE_RQ
            The received C-STORE-RQ message.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        LOGGER.info('Received Store Request')

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-STORE RQ'))
        s.append('Presentation Context ID       : {0!s}'
                 .format(msg.context_id))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Affected SOP Instance UID     : {0!s}'
                 .format(cs.AffectedSOPInstanceUID))
        if 'MoveOriginatorApplicationEntityTitle' in cs:
            s.append('Move Originator               : {0!s}'
                     .format(cs.MoveOriginatorApplicationEntityTitle))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('Priority                      : {0!s}'.format(priority))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_receive_c_store_rsp(msg):
        """Debugging function when a C-STORE-RSP is received.

        **C-STORE Response Elements**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Affected SOP Instance UID
        | (M) Status

        Parameters
        ----------
        msg : dimse_messagesC_STORE_RSP
            The received C-STORE-RSP message.
        """
        cs = msg.command_set

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        # See PS3.4 Annex B.2.3 for Storage Service Class Statuses
        status_str = '0x{0:04x} - Unknown'.format(cs.Status)
        # Try and get the status from the affected SOP class UID
        if 'AffectedSOPClassUID' in cs:
            service_class = uid_to_service_class(cs.AffectedSOPClassUID)
            try:
                if cs.Status in service_class.statuses:
                    status = service_class.statuses[cs.Status]
                    status_str = '0x{0:04x} - {1}'.format(cs.Status, status[0])
            except AttributeError:
                status_str = '0x{0:04x}'.format(cs.Status)

        LOGGER.info('Received Store Response')
        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-STORE RSP'))
        s.append('Presentation Context ID       : {0!s}'
                 .format(msg.context_id))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'AffectedSOPInstanceUID' in cs:
            s.append('Affected SOP Instance UID     : {0!s}'
                     .format(cs.AffectedSOPInstanceUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('DIMSE Status                  : {0!s}'.format(status_str))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_receive_c_find_rq(msg):
        """Debugging function when a C-FIND-RQ is received.

        **C-FIND Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Priority
        | (M) Identifier

        Parameters
        ----------
        msg : dimse_messagesC_FIND_RQ
            The received C-FIND-RQ message.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-FIND RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('Priority                      : {0!s}'.format(priority))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.info(line)

    @staticmethod
    def debug_receive_c_find_rsp(msg):
        """Debugging function when a C-FIND-RSP is received.

        **C-FIND Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (C) Identifier
        | (M) Status

        Parameters
        ----------
        msg : dimse_messages.C_FIND_RSP
            The received C-FIND-RSP message.
        """
        cs = msg.command_set
        if cs.Status != 0x0000:
            return

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-FIND RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('DIMSE Status                  : 0x{0:04x}'.format(cs.Status))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.info(line)

    @staticmethod
    def debug_receive_c_cancel_rq(msg):
        """Debugging function when a C-CANCEL-RQ is received.

        Covers C-CANCEL-FIND-RQ, C-CANCEL-GET-RQ and C-CANCEL-MOVE-RQ

        Parameters
        ----------
        msg : dimse_messagesC_CANCEL_RQ
            The received C-CANCEL-RQ message.
        """
        cs = msg.command_set

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-CANCEL RQ'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.info(line)

    @staticmethod
    def debug_receive_c_get_rq(msg):
        """Debugging function when a C-GET-RQ is received.

        **C-GET Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Priority
        | (M) Identifier

        Parameters
        ----------
        msg : dimse_messagesC_GET_RQ
            The received C-GET-RQ message.
        """
        cs = msg.command_set

        priority_str = {2: 'Low',
                        0: 'Medium',
                        1: 'High'}
        priority = priority_str[cs.Priority]

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-GET RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('Priority                      : {0!s}'.format(priority))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.info(line)

    @staticmethod
    def debug_receive_c_get_rsp(msg):
        """Debugging function when a C-GET-RSP is received.

        **C-GET Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Identifier
        | (M) Status
        | (C) Number of Remaining Sub-operations
        | (C) Number of Completed Sub-operations
        | (C) Number of Failed Sub-operations
        | (C) Number of Warning Sub-operations

        Parameters
        ----------
        msg : dimse_messagesC_GET_RSP
            The received C-GET-RSP message.
        """
        cs = msg.command_set

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-GET RSP'))
        s.append('Presentation Context ID       : {0!s}'
                 .format(msg.context_id))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'NumberOfRemainingSuboperations' in cs:
            s.append('Remaining Sub-operations      : {0!s}'
                     .format(cs.NumberOfRemainingSuboperations))
        if 'NumberOfCompletedSuboperations' in cs:
            s.append('Completed Sub-operations      : {0!s}'
                     .format(cs.NumberOfCompletedSuboperations))
        if 'NumberOfFailedSuboperations' in cs:
            s.append('Failed Sub-operations         : {0!s}'
                     .format(cs.NumberOfFailedSuboperations))
        if 'NumberOfWarningSuboperations' in cs:
            s.append('Warning Sub-operations        : {0!s}'
                     .format(cs.NumberOfWarningSuboperations))
        s.append('Data Set                      : {0!s}'.format(dataset))
        s.append('DIMSE Status                  : 0x{0:04x}'.format(cs.Status))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_receive_c_move_rq(msg):
        """Debugging function when a C-MOVE-RQ is received.

        **C-MOVE Request Parameters**

        | (M) Message ID
        | (M) Affected SOP Class UID
        | (M) Priority
        | (M) Move Destination
        | (M) Identifier

        Parameters
        ----------
        msg : dimse_messagesC_MOVE_RQ
            The received C-MOVE-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_c_move_rsp(msg):
        """Debugging function when a C-MOVE-RSP is received.

        **C-MOVE Response Parameters**

        | (U) Message ID
        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Identifier
        | (M) Status
        | (C) Number of Remaining Sub-operations
        | (C) Number of Completed Sub-operations
        | (C) Number of Failed Sub-operations
        | (C) Number of Warning Sub-operations

        Parameters
        ----------
        msg : dimse_messagesC_MOVE_RSP
            The received C-MOVE-RSP message.
        """
        cs = msg.command_set

        identifier = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            identifier = 'Present'

        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('C-MOVE RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'NumberOfRemainingSuboperations' in cs:
            s.append('Remaining Sub-operations      : {0!s}'
                     .format(cs.NumberOfRemainingSuboperations))
        if 'NumberOfCompletedSuboperations' in cs:
            s.append('Completed Sub-operations      : {0!s}'
                     .format(cs.NumberOfCompletedSuboperations))
        if 'NumberOfFailedSuboperations' in cs:
            s.append('Failed Sub-operations         : {0!s}'
                     .format(cs.NumberOfFailedSuboperations))
        if 'NumberOfWarningSuboperations' in cs:
            s.append('Warning Sub-operations        : {0!s}'
                     .format(cs.NumberOfWarningSuboperations))
        s.append('Identifier                    : {0!s}'.format(identifier))
        s.append('DIMSE Status                  : 0x{0:04x}'.format(cs.Status))

        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_event_report_rq(msg):
        """Debugging function when an N-EVENT-REPORT-RQ is sent.

        Parameters
        ----------
        msg : dimse_messages.N_EVENT_REPORT_RQ
            The N-EVENT-REPORT-RQ message to be sent.
        """
        cs = msg.command_set

        evt_info = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            evt_info = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'
                 .format('N-EVENT-REPORT RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Affected SOP Class UID        : {0!s}'
                 .format(cs.AffectedSOPClassUID))
        s.append('Affected SOP Instance UID     : {0!s}'
                 .format(cs.AffectedSOPInstanceUID))
        s.append('Event Type ID                 : {0!s}'
                 .format(cs.EventTypeID))
        s.append('Event Information             : {0!s}'.format(evt_info))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_event_report_rsp(msg):
        """Debugging function when an N-EVENT-REPORT-RSP is sent.

        Parameters
        ----------
        msg : dimse_messages.N_EVENT_REPORT_RSP
            The N-EVENT-REPORT-RSP message to be sent.
        """
        cs = msg.command_set

        evt_reply = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            evt_reply = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'
                 .format('N-EVENT-REPORT RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'AffectedSOPInstanceUID' in cs:
            s.append('Affected SOP Instance UID     : {0!s}'
                     .format(cs.AffectedSOPInstanceUID))
        if 'EventTypeID' in cs:
            s.append(
                'Event Type ID                 : {!s}'
                .format(cs.EventTypeID)
            )
        s.append('Event Reply                       : {0!s}'.format(evt_reply))
        s.append('Status                        : 0x{0:04x}'.format(cs.Status))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_get_rq(msg):
        """Debugging function when an N-GET-RQ is sent.

        Parameters
        ----------
        msg : dimse_messages.N_GET_RQ
            The N-GET-RQ message to be sent.
        """
        cs = msg.command_set

        nr_attr = 'no identifiers'
        if 'AttributeIdentifierList' in cs:
            nr_attr = len(cs.AttributeIdentifierList)
            if nr_attr == 1:
                nr_attr = '{} identifier'.format(nr_attr)
            else:
                nr_attr = '{} identifiers'.format(nr_attr)

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-GET RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Requested SOP Class UID       : {0!s}'
                 .format(cs.RequestedSOPClassUID))
        s.append('Requested SOP Instance UID    : {0!s}'
                 .format(cs.RequestedSOPInstanceUID))
        s.append('Attribute Identifier List     : ({0!s})'.format(nr_attr))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_get_rsp(msg):
        """Debugging function when an N-GET-RSP is sent.

        Parameters
        ----------
        msg : dimse_messages.N_GET_RSP
            The N-GET-RSP message to be sent.
        """
        cs = msg.command_set

        attr_list = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            attr_list = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-GET RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'AffectedSOPInstanceUID' in cs:
            s.append('Affected SOP Instance UID     : {0!s}'
                     .format(cs.AffectedSOPInstanceUID))
        s.append('Attribute List                : {0!s}'.format(attr_list))
        s.append('Status                        : 0x{0:04x}'.format(cs.Status))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_set_rq(msg):
        """Debugging function when an N-SET-RQ is sent.

        Parameters
        ----------
        msg : dimse_messages.N_SET_RQ
            The N-SET-RQ message to be sent.
        """
        cs = msg.command_set

        mod_list = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            mod_list = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-SET RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Requested SOP Class UID       : {0!s}'
                 .format(cs.RequestedSOPClassUID))
        s.append('Requested SOP Instance UID    : {0!s}'
                 .format(cs.RequestedSOPInstanceUID))
        s.append('Modification List             : {0!s}'.format(mod_list))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_set_rsp(msg):
        """Debugging function when an N-SET-RSP is sent.

        Parameters
        ----------
        msg : dimse_messages.N_SET_RSP
            The N-SET-RSP message to be sent.
        """
        cs = msg.command_set

        attr_list = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            attr_list = 'Present'

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-SET RSP'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'AffectedSOPInstanceUID' in cs:
            s.append('Affected SOP Instance UID     : {0!s}'
                     .format(cs.AffectedSOPInstanceUID))
        s.append('Attribute List                : {0!s}'.format(attr_list))
        s.append('Status                        : 0x{0:04x}'.format(cs.Status))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_action_rq(msg):
        """Debugging function when an N-ACTION-RQ is sent.

        Parameters
        ----------
        msg : dimse_messages.N_ACTION_RQ
            The N-ACTION-RQ message to be sent.
        """
        pass

    @staticmethod
    def debug_send_n_action_rsp(msg):
        """Debugging function when an N-ACTION-RSP is sent.

        Parameters
        ----------
        msg : dimse_messages.N_ACTION_RSP
            The N-ACTION-RSP message to be sent.
        """
        pass

    @staticmethod
    def debug_send_n_create_rq(msg):
        """Debugging function when an N-CREATE-RQ is sent.

        Parameters
        ----------
        msg : dimse_messages.N_CREATE_RQ
            The N-CREATE-RQ message to be sent.
        """
        pass

    @staticmethod
    def debug_send_n_create_rsp(msg):
        """Debugging function when an N-CREATE-RSP is sent.

        Parameters
        ----------
        msg : dimse_messages.N_CREATE_RSP
            The N-CREATE-RSP message to be sent.
        """
        pass

    @staticmethod
    def debug_send_n_delete_rq(msg):
        """Debugging function when an N-DELETE-RQ is sent.

        Parameters
        ----------
        msg : dimse_messages.N_DELETE_RQ
            The N-DELETE-RQ message to be sent.
        """
        cs = msg.command_set

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-DELETE RQ'))
        s.append('Message ID                    : {0!s}'.format(cs.MessageID))
        s.append('Requested SOP Class UID       : {0!s}'
                 .format(cs.RequestedSOPClassUID))
        s.append('Requested SOP Instance UID    : {0!s}'
                 .format(cs.RequestedSOPInstanceUID))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_send_n_delete_rsp(msg):
        """Debugging function when an N-DELETE-RSP is sent.

        Parameters
        ----------
        msg : dimse_messages.N_DELETE_RSP
            The N-DELETE-RSP message to be sent.
        """
        cs = msg.command_set

        s = []
        s.append(
            '===================== OUTGOING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-DELETE RQ'))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'AffectedSOPInstanceUID' in cs:
            s.append('Affected SOP Instance UID     : {0!s}'
                     .format(cs.AffectedSOPInstanceUID))
        s.append('Status                        : 0x{0:04x}'.format(cs.Status))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )
        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_receive_n_event_report_rq(msg):
        """Debugging function when an N-EVENT-REPORT-RQ is received.

        Parameters
        ----------
        msg : dimse_messages.N_EVENT_REPORT_RQ
            The received N-EVENT-REPORT-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_n_event_report_rsp(msg):
        """Debugging function when an N-EVENT-REPORT-RSP is received.

        Parameters
        ----------
        msg : dimse_messages.N_EVENT_REPORT_RSP
            The received N-EVENT-REPORT-RSP message.
        """
        pass

    @staticmethod
    def debug_receive_n_get_rq(msg):
        """Debugging function when an N-GET-RQ is received.

        Parameters
        ----------
        msg : dimse_messages.N_GET_RQ
            The received N-GET-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_n_get_rsp(msg):
        """Debugging function when an N-GET-RSP is received.

        Parameters
        ----------
        msg : dimse_messages.N_GET_RSP
            The received N-GET-RSP message.
        """
        cs = msg.command_set

        dataset = 'None'
        if msg.data_set and msg.data_set.getvalue() != b'':
            dataset = 'Present'

        LOGGER.info('Received Get Response')
        s = []
        s.append(
            '===================== INCOMING DIMSE MESSAGE ===================='
        )
        s.append('Message Type                  : {0!s}'.format('N-GET RSP'))
        s.append('Presentation Context ID       : {0!s}'
                 .format(msg.context_id))
        s.append('Message ID Being Responded To : {0!s}'
                 .format(cs.MessageIDBeingRespondedTo))
        if 'AffectedSOPClassUID' in cs:
            s.append('Affected SOP Class UID        : {0!s}'
                     .format(cs.AffectedSOPClassUID))
        if 'AffectedSOPInstanceUID' in cs:
            s.append('Affected SOP Instance UID     : {0!s}'
                     .format(cs.AffectedSOPInstanceUID))
        s.append('Attribute List                : {0!s}'.format(dataset))
        s.append('Status                        : 0x{0:04x}'.format(cs.Status))
        s.append(
            '======================= END DIMSE MESSAGE ======================='
        )

        for line in s:
            LOGGER.debug(line)

    @staticmethod
    def debug_receive_n_set_rq(msg):
        """Debugging function when an N-SET-RQ is received.

        Parameters
        ----------
        msg : dimse_messages.N_SET_RQ
            The received N-SET-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_n_set_rsp(msg):
        """Debugging function when an N-SET-RSP is received.

        Parameters
        ----------
        msg : dimse_messages.N_SET_RSP
            The received N-SET-RSP message.
        """
        pass

    @staticmethod
    def debug_receive_n_action_rq(msg):
        """Debugging function when an N-ACTION-RQ is received.

        Parameters
        ----------
        msg : dimse_messages.N_ACTION_RQ
            The received N-ACTION-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_n_action_rsp(msg):
        """Debugging function when an N-ACTION-RSP is received.

        Parameters
        ----------
        msg : dimse_messages.N_ACTION_RSP
            The received N-ACTION-RSP message.
        """
        pass

    @staticmethod
    def debug_receive_n_create_rq(msg):
        """Debugging function when an N-CREATE-RQ is received.

        Parameters
        ----------
        msg : dimse_messages.N_CREATE_RQ
            The received N-CREATE-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_n_create_rsp(msg):
        """Debugging function when an N-CREATE-RSP is received.

        Parameters
        ----------
        msg : dimse_messages.N_CREATE_RSP
            The received N-CREATE-RSP message.
        """
        pass

    @staticmethod
    def debug_receive_n_delete_rq(msg):
        """Debugging function when an N-DELETE-RQ is received.

        Parameters
        ----------
        msg : dimse_messages.N_DELETE_RQ
            The received N-DELETE-RQ message.
        """
        pass

    @staticmethod
    def debug_receive_n_delete_rsp(msg):
        """Debugging function when an N-DELETE-RSP is received.

        Parameters
        ----------
        msg : dimse_messages.N_DELETE_RSP
            The received N-DELETE-RSP message.
        """
        pass
